"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Request = void 0;
const https_1 = require("https");
const http_1 = require("http");
const zlib_1 = require("zlib");
const querystring_1 = require("querystring");
const url_1 = require("url");
const stream_1 = require("stream");
const base_client_1 = require("./base-client");
const request_error_1 = require("./request-error");
const protos = { http: http_1.request, https: https_1.request }, ports = { http: 80, https: 443 };
class Request extends base_client_1.BaseClient {
    constructor(method, url) {
        super();
        this._body = '';
        const parsed = (0, url_1.parse)(url);
        this._protocol = (parsed.protocol || 'http:').slice(0, -1);
        this._query = (0, querystring_1.parse)(parsed.query);
        this._params = {
            host: parsed.hostname,
            port: parsed.port || ports[this._protocol],
            path: parsed.pathname,
            method: method,
        };
        this._headers['Accept'] = 'application/json';
        this._headers['Accept-Encoding'] = 'gzip, deflate, br';
    }
    query(arg) {
        let inner;
        if (typeof arg === 'string') {
            inner = (0, querystring_1.parse)(arg);
        }
        else {
            inner = arg;
        }
        Object.keys(inner).forEach((k) => this._query[k] = inner[k]);
        return this;
    }
    cert(certstr) {
        this._params.cert = certstr;
        return this;
    }
    key(keystr) {
        this._params.key = keystr;
        return this;
    }
    ca(castrs) {
        this._params.ca = castrs;
        return this;
    }
    send(body) {
        if (typeof body === 'string') {
            this._body = body;
        }
        else {
            this._body = JSON.stringify(body);
            this._headers['Content-Type'] = 'application/json';
        }
        this._headers['Content-Length'] = Buffer.byteLength(this._body);
        return this;
    }
    attempt() {
        let req;
        const params = Object.assign({}, this._params);
        if (Object.keys(this._query).length) {
            params.path += '?' + (0, querystring_1.stringify)(Object.assign({}, this._query));
        }
        params.headers = Object.assign({}, this._headers);
        const promise = new Promise((resolve, reject) => {
            let res;
            req = protos[this._protocol](params, (response) => {
                res = response;
                res.text = '';
                const output = new stream_1.Writable();
                const onError = (err) => {
                    if (err)
                        console.log(err);
                    output.end();
                };
                output._write = (chunk, encoding, next) => {
                    res.text += chunk.toString();
                    next();
                };
                output.on('close', () => {
                    try {
                        res.body = JSON.parse(res.text);
                    }
                    catch (e) {
                        res.body = null;
                    }
                    if (res.statusCode >= 300) {
                        const rErr = new request_error_1.RequestError('Bad response from server.', this, params, res);
                        return reject(rErr);
                    }
                    resolve(res);
                });
                switch (res.headers['content-encoding']) {
                    case 'br':
                        (0, stream_1.pipeline)(res, (0, zlib_1.createBrotliDecompress)(), output, onError);
                        break;
                    case 'gzip':
                    case 'deflate':
                        (0, stream_1.pipeline)(res, (0, zlib_1.createUnzip)(), output, onError);
                        break;
                    default:
                        (0, stream_1.pipeline)(res, output, onError);
                        break;
                }
            });
            if (this._body.length)
                req.write(this._body);
            req.on('error', (err) => {
                const rErr = new request_error_1.RequestError('Connection Error: ' + err.message, this, params);
                reject(rErr);
            });
            req.on('abort', () => {
                const rErr = new request_error_1.RequestError(`Request was aborted`, this, params);
                reject(rErr);
            });
            req.end();
        });
        if (this._ttl) {
            return Promise.race([
                promise,
                new Promise((_, r) => {
                    setTimeout(() => {
                        req.abort();
                        const error = new request_error_1.RequestError('Request timed out', this, params);
                        r(error);
                    }, this._ttl);
                })
            ]);
        }
        else {
            return promise;
        }
    }
    async promise() {
        let attempts = 0;
        while (true) {
            try {
                return await this.attempt();
            }
            catch (e) {
                if ((e.statusCode >= 500 || !e.statusCode) && this._retry.retries > attempts) {
                    attempts += 1;
                    if (this._retry.delay) {
                        let wait = this._retry.delay;
                        if (this._retry.backoff) {
                            wait *= this._retry.backoff ** (attempts - 1);
                        }
                        await new Promise(r => setTimeout(r, wait));
                    }
                }
                else {
                    throw e;
                }
            }
        }
    }
    /*
     * Promise-like functions
     */
    then(resolve, reject) {
        return this.promise().then(resolve, reject);
    }
    catch(cb) {
        return this.promise().catch(cb);
    }
    finally(cb) {
        return this.promise().finally(cb);
    }
    /*
     * Callback support
     */
    end(done) {
        this.then((res) => done(null, res), (err) => done(err));
    }
}
exports.Request = Request;
